#include "camera.h"
#include "quaternion.h"
#include <stdio.h>

#ifdef _WIN32
#include <windows.h>
#define GLUT_DISABLE_ATEXIT_HACK
#endif

#ifdef __APPLE__
#include <OpenGL/glu.h>
#else
#include <GL/glu.h>
#endif

UserCamera::UserCamera() : 
//    mDesiredOrientation(cos(PI/4), sin(PI/4), 0, 0),
    mDesiredOrientation(1, 0, 0, 0),
    mCurrentOrientation(1, 0, 0, 0),
    mUp(0, 1, 0),
    mRight(1, 0, 0),
    mForward(0, 0, -1),
    mPos(0, 1, 0),
    mSpeed(4),
    mMovingForward(false),
    mMovingBackward(false),
    mMovingRight(false),
    mMovingLeft(false),
    mMovingUp(false),
    mMovingDown(false)
{
}

UserCamera::~UserCamera() {
}

void UserCamera::beginForward() {
    mMovingForward = true;
}

void UserCamera::beginBackward() {
    mMovingBackward = true;
}

void UserCamera::beginLeft() {
    mMovingLeft = true;
}

void UserCamera::beginRight() {
    mMovingRight = true;
}
void UserCamera::beginUp() {
    mMovingUp = true;
}
void UserCamera::beginDown() {
    mMovingDown = true;
}

void UserCamera::endForward() {
    mMovingForward = false;
}

void UserCamera::endBackward() {
    mMovingBackward = false;
}

void UserCamera::endLeft() {
    mMovingLeft = false;
}

void UserCamera::endRight() {
    mMovingRight = false;
}

void UserCamera::endUp() {
    mMovingUp = false;
}
void UserCamera::endDown() {
    mMovingDown = false;
}

void UserCamera::rotateAboutRight(double angle) {
    vector3 right = QVRotate(mDesiredOrientation, mRight);
    quaternion q(cos(angle/2), sin(angle/2)*right);

    mDesiredOrientation = q*mDesiredOrientation;
}

void UserCamera::rotateAboutUp(double angle) {
    vector3 up = QVRotate(mDesiredOrientation, mUp);
    quaternion q(cos(angle/2), sin(angle/2)*up);

    mDesiredOrientation = q*mDesiredOrientation;
}

void UserCamera::rotateAboutForward(double angle) {
    vector3 forward = QVRotate(mDesiredOrientation, mForward);
    quaternion q(cos(angle/2), sin(angle/2)*forward);

    mDesiredOrientation = q*mDesiredOrientation;
}

void UserCamera::update(double dt) {
    vector3 forward = QVRotate(mDesiredOrientation, mForward);
    vector3 right = QVRotate(mDesiredOrientation, mRight);
    vector3 up = QVRotate(mDesiredOrientation, mUp);

    vector3 move;
    if(mMovingForward) {
    	move += forward;
    }

    if(mMovingBackward) {
    	move += -forward;
    }

    if(mMovingRight) {
    	move += right;
    }

    if(mMovingLeft) {
    	move += -right;
    }

    if(mMovingUp) {
    	move += up;
    }

    if(mMovingDown) {
    	move += -up;
    }

    double length = move.length();

    if(length > 0) {
        move /= length;
        move *= mSpeed;
    }

    mPos += move*dt;

    mCurrentOrientation = slerp(mCurrentOrientation, mDesiredOrientation, .3);
    mCurrentOrientation.normalize();
}

void UserCamera::apply() {
    vector3 forward = QVRotate(mCurrentOrientation, mForward);
    vector3 up = QVRotate(mCurrentOrientation, mUp);
    vector3 at = mPos+forward;
    gluLookAt(mPos.x, mPos.y, mPos.z,
              at.x, at.y, at.z,
              up.x, up.y, up.z);
}

void UserCamera::debug() {
    printf("%f %f %f %f\n", mCurrentOrientation.n, mCurrentOrientation.v.x, mCurrentOrientation.v.y, mCurrentOrientation.v.z);
    printf("%f %f %f %f\n", mDesiredOrientation.n, mDesiredOrientation.v.x, mDesiredOrientation.v.y, mDesiredOrientation.v.z);
}

vector3 UserCamera::getPosition() {
    return mPos;
}
