#ifndef MATRIX3X3
#define MATRIX3X3

#include "vector3.h"

class matrix3x3 {
    public:
        double e11, e12, e13,
               e21, e22, e23,
               e31, e32, e33;

        matrix3x3() {
            e11 = e12 = e13 = e21 = e22 = e23 = e31 = e32 = e33 = 0;
        }

        matrix3x3(double r1c1, double r1c2, double r1c3,
                  double r2c1, double r2c2, double r2c3,
                  double r3c1, double r3c2, double r3c3) {
            e11 = r1c1;
            e12 = r1c2;
            e13 = r1c3;

            e21 = r2c1;
            e22 = r2c2;
            e23 = r2c3;

            e31 = r3c1;
            e32 = r3c2;
            e33 = r3c3;
        }

        inline double det() {
            return e11*e22*e33-
                   e11*e32*e23+
                   e21*e32*e13-
                   e21*e12*e33+
                   e31*e12*e23-
                   e31*e22*e13;
        }
        inline matrix3x3 Transpose() {
            return matrix3x3(e11, e21, e31, e12, e22, e32, e13, e23, e33);
        }

        inline matrix3x3 Inverse() {
            double d = e11*e22*e33-
                   e11*e32*e23+
                   e21*e32*e13-
                   e21*e12*e33+
                   e31*e12*e23-
                   e31*e22*e13;
            if(d == 0) d = 1;

            return matrix3x3( (e22*e33-e23*e32)/d,
                              -(e12*e33-e13*e32)/d,
                              (e12*e23-e13*e22)/d,
                              -(e21*e33-e23*e31)/d,
                              (e11*e33-e13*e31)/d,
                              -(e11*e23-e13*e21)/d,
                              (e21*e32-e22*e31)/d,
                              -(e11*e32-e12*e31)/d,
                              (e11*e22-e12*e21)/d );
        }

        inline void copy_matrix(double* m) {
            m[0] = e11;
            m[1] = e21;
            m[2] = e31;
            m[3] = 0;
            m[4] = e12;
            m[5] = e22;
            m[6] = e32;
            m[7] = 0;
            m[8] = e13;
            m[9] = e23;
            m[10] = e33;
            m[11] = 0;
            m[12] = 0;
            m[13] = 0;
            m[14] = 0;
            m[15] = 1;
        }

        inline matrix3x3& operator+=(matrix3x3 m) {
            e11 += m.e11;
            e12 += m.e12;
            e13 += m.e13;
            e21 += m.e21;
            e22 += m.e22;
            e23 += m.e23;
            e31 += m.e31;
            e32 += m.e32;
            e33 += m.e33;
            return *this;
        }

        inline matrix3x3& operator-=(matrix3x3 m) {
            e11 -= m.e11;
            e12 -= m.e12;
            e13 -= m.e13;
            e21 -= m.e21;
            e22 -= m.e22;
            e23 -= m.e23;
            e31 -= m.e31;
            e32 -= m.e32;
            e33 -= m.e33;
            return *this;
        }

        inline matrix3x3& operator*=(double s) {
            e11 *= s;
            e12 *= s;
            e13 *= s;
            e21 *= s;
            e22 *= s;
            e23 *= s;
            e31 *= s;
            e32 *= s;
            e33 *= s;
            return *this;
        }

        inline matrix3x3& operator/=(double s) {
            e11 /= s;
            e12 /= s;
            e13 /= s;
            e21 /= s;
            e22 /= s;
            e23 /= s;
            e31 /= s;
            e32 /= s;
            e33 /= s;
            return *this;
        }
};

inline matrix3x3 operator+(matrix3x3 m1, matrix3x3 m2) {
    return matrix3x3(m1.e11+m2.e11,
                     m1.e12+m2.e12,
                     m1.e13+m2.e13,
                     m1.e21+m2.e21,
                     m1.e22+m2.e22,
                     m1.e23+m2.e23,
                     m1.e31+m2.e31,
                     m1.e32+m2.e32,
                     m1.e33+m2.e33);
}

inline matrix3x3 operator-(matrix3x3 m1, matrix3x3 m2) {
    return matrix3x3(m1.e11-m2.e11,
                     m1.e12-m2.e12,
                     m1.e13-m2.e13,
                     m1.e21-m2.e21,
                     m1.e22-m2.e22,
                     m1.e23-m2.e23,
                     m1.e31-m2.e31,
                     m1.e32-m2.e32,
                     m1.e33-m2.e33);
}

inline matrix3x3 operator/(matrix3x3 m, double s) {
    return matrix3x3(m.e11/s,
                     m.e12/s,
                     m.e13/s,
                     m.e21/s,
                     m.e22/s,
                     m.e23/s,
                     m.e31/s,
                     m.e32/s,
                     m.e33/s);
}

inline matrix3x3 operator*(matrix3x3 m1, matrix3x3 m2) {
    return matrix3x3(m1.e11*m2.e11 + m1.e12*m2.e21 + m1.e13*m2.e31,
                     m1.e11*m2.e12 + m1.e12*m2.e22 + m1.e13*m2.e32,
                     m1.e11*m2.e13 + m1.e12*m2.e23 + m1.e13*m2.e33,
                     m1.e21*m2.e11 + m1.e22*m2.e21 + m1.e23*m2.e31,
                     m1.e21*m2.e12 + m1.e22*m2.e22 + m1.e23*m2.e32,
                     m1.e21*m2.e13 + m1.e22*m2.e23 + m1.e23*m2.e33,
                     m1.e31*m2.e11 + m1.e32*m2.e21 + m1.e33*m2.e31,
                     m1.e31*m2.e12 + m1.e32*m2.e22 + m1.e33*m2.e32,
                     m1.e31*m2.e13 + m1.e32*m2.e23 + m1.e33*m2.e33);
}

inline matrix3x3 operator*(matrix3x3 m, double s) {
    return matrix3x3(m.e11*s,
                     m.e12*s,
                     m.e13*s,
                     m.e21*s,
                     m.e22*s,
                     m.e23*s,
                     m.e31*s,
                     m.e32*s,
                     m.e33*s);
}


inline matrix3x3 operator*(double s, matrix3x3 m) {
    return matrix3x3(m.e11*s,
                     m.e12*s,
                     m.e13*s,
                     m.e21*s,
                     m.e22*s,
                     m.e23*s,
                     m.e31*s,
                     m.e32*s,
                     m.e33*s);
}

inline vector3 operator*(matrix3x3 m, vector3 v) {
    return vector3(m.e11*v.x+m.e12*v.y+m.e13*v.z,
                   m.e21*v.x+m.e22*v.y+m.e23*v.z,
                   m.e31*v.x+m.e32*v.y+m.e33*v.z);
}

inline vector3 operator*(vector3 v, matrix3x3 m) {
    return vector3(m.e11*v.x+m.e21*v.y+m.e31*v.z,
                   m.e12*v.x+m.e22*v.y+m.e32*v.z,
                   m.e13*v.x+m.e23*v.y+m.e33*v.z);
}

inline matrix3x3 identity_matrix3x3() {
    return matrix3x3(1,0,0,
                     0,1,0,
                     0,0,1);
}

inline matrix3x3 Star(vector3 v) {
    return matrix3x3(0, -v.z, v.y,
                     v.z, 0, -v.x,
                     -v.y, v.x, 0);
}

#endif
