#ifndef QUATERNION
#define QUATERNION
#include <stdio.h>

#include "matrix3x3.h"
#include "vector3.h"
#include <math.h>

#define QUATERNION_TOL .0001
#ifndef PI
#define PI 3.1415926535
#endif

inline double DegreesToRadians(double deg) {
    return deg * PI / 180.0;
}

inline double RadiansToDegrees(double rad) {
    return rad * 180.0 / PI;
}

class quaternion {
public:
    double n;
    vector3 v;

    quaternion() {
        n = 1;
        v.x = 0;
        v.y = 0;
        v.z = 0;
    }

    quaternion(double nn, vector3 vv) : n(nn), v(vv) {
    }

    quaternion(double e0, double e1, double e2, double e3) {
        n = e0;
        v.x = e1;
        v.y = e2;
        v.z = e3;
    }

    inline double magnitude() {
        return sqrt(n*n + v.x*v.x + v.y*v.y + v.z*v.z);
    }

    inline vector3 GetVector() {
        return vector3(v.x, v.y, v.z);
    }

    inline double GetScalar() {
        return n;
    }

    inline quaternion conjugate() {
        return quaternion(n, -v.x, -v.y, -v.z);
    }

    inline quaternion operator+=(quaternion q) {
        n += q.n;
        v.x += q.v.x;
        v.y += q.v.y;
        v.z += q.v.z;

        return *this;
    }

    inline quaternion operator-=(quaternion q) {
        n -= q.n;
        v.x -= q.v.x;
        v.y -= q.v.y;
        v.z -= q.v.z;

        return *this;
    }

    inline quaternion operator*=(double s) {
        n *= s;
        v.x *= s;
        v.y *= s;
        v.z *= s;

        return *this;
    }

    inline quaternion operator*=(quaternion q) {

        double tmpn = n*q.n - v.x*q.v.x - v.y*q.v.y - v.z*q.v.z;
        double tmpx = n*q.v.x + v.x*q.n + v.y*q.v.z - v.z*q.v.y;
        double tmpy = n*q.v.y + v.y*q.n + v.z*q.v.x - v.x*q.v.z;
        double tmpz = n*q.v.z + v.z*q.n + v.x*q.v.y - v.y*q.v.x;

        n = tmpn;
        v.x = tmpx;
        v.y = tmpy;
        v.z = tmpz;

        return *this;
    }

    inline quaternion operator/=(double s) {
        n /= s;
        v.x /= s;
        v.y /= s;
        v.z /= s;

        return *this;
    }

    quaternion operator~() const { return quaternion(n, -v.x, -v.y, -v.z); }

    inline quaternion& normalize() {
        double m = sqrt(n*n + v.x*v.x + v.y*v.y + v.z*v.z);

        if(m > QUATERNION_TOL) {
            n /= m;
            v /= m;
        }
        return *this;
    }
};

inline quaternion operator+(quaternion q1, quaternion q2) {
    return quaternion(q1.n+q2.n,
                      q1.v.x + q2.v.x,
                      q1.v.y + q2.v.y,
                      q1.v.z + q2.v.z);
}

inline quaternion operator-(quaternion q1, quaternion q2) {
    return quaternion(q1.n-q2.n,
                      q1.v.x - q2.v.x,
                      q1.v.y - q2.v.y,
                      q1.v.z - q2.v.z);
}

inline quaternion operator*(quaternion q1, quaternion q2) {
    return quaternion(q1.n*q2.n - q1.v.x*q2.v.x - q1.v.y*q2.v.y - q1.v.z*q2.v.z,
                      q1.n*q2.v.x + q1.v.x*q2.n + q1.v.y*q2.v.z - q1.v.z*q2.v.y,
                      q1.n*q2.v.y + q1.v.y*q2.n + q1.v.z*q2.v.x - q1.v.x*q2.v.z,
                      q1.n*q2.v.z + q1.v.z*q2.n + q1.v.x*q2.v.y - q1.v.y*q2.v.x);
}

inline quaternion operator*(quaternion q, double s) {
    return quaternion(q.n*s, q.v.x*s, q.v.y*s, q.v.z*s);
}

inline quaternion operator*(double s, quaternion q) {
    return quaternion(q.n*s, q.v.x*s, q.v.y*s, q.v.z*s);
}

inline quaternion slerp(quaternion q1, quaternion q2, double t) {
    double inner_product = q1.n*q2.n + q1.v.x*q2.v.x + q1.v.y*q2.v.y + q1.v.z*q2.v.z;

    if(inner_product > 1) {
    	inner_product = 1;
    } else if(inner_product < -1) {
    	inner_product = -1;
    }

    double phi = acos(inner_product);
    double one_minus_t = 1-t;
    if(phi < .01) {
    	return one_minus_t*q1+t*q2;
    } else {
    	double m1 = sin(phi*one_minus_t)/sin(phi);
    	double m2 = sin(phi*t)/sin(phi);

    	return m1*q1+m2*q2;
    }
}

inline quaternion operator*(quaternion q, vector3 v) {
    return quaternion( -(q.v.x*v.x + q.v.y*v.y + q.v.z*v.z),
                        q.n*v.x + q.v.y*v.z - q.v.z*v.y,
                        q.n*v.y + q.v.z*v.x - q.v.x*v.z,
                        q.n*v.z + q.v.x*v.y - q.v.y*v.x);
}

inline quaternion operator*(vector3 v, quaternion q) {
    return quaternion( -(q.v.x*v.x + q.v.y*v.y + q.v.z*v.z),
                       q.n*v.x + q.v.z*v.y - q.v.y*v.z,
                       q.n*v.y + q.v.x*v.z - q.v.z*v.x,
                       q.n*v.z + q.v.y*v.x - q.v.x*v.y);
}

inline quaternion operator/(quaternion q, double s) {
    return quaternion(q.n/s, q.v.x/s, q.v.y/s, q.v.z/s);
}

inline double QGetAngle(quaternion q) {
    return 2*acos(q.n);
}

inline vector3 QGetAxis(quaternion q) {
    vector3 v;
    double m;

    v = q.GetVector();
    m = v.magnitude();

    if(m <= QUATERNION_TOL) return vector3();
    else return v/m;
}

inline quaternion QRotate(quaternion q1, quaternion q2) {
    return q1*q2*(~q1);
}

inline vector3 QVRotate(quaternion q, vector3 v) {
    quaternion t;

    t = q*v*(~q);
    return t.GetVector();
}

inline quaternion MakeQFromEulerAngles(double x, double y, double z) {
    quaternion q;
    double roll = DegreesToRadians(x);
    double pitch = DegreesToRadians(y);
    double yaw = DegreesToRadians(z);

    double cyaw, cpitch, croll, syaw, spitch, sroll;
    double cyawcpitch, syawspitch, cyawspitch, syawcpitch;

    cyaw = cos(0.5f*yaw);
    cpitch = cos(0.5f*pitch);
    croll = cos(0.5f*roll);
    syaw = sin(0.5f*yaw);
    spitch = sin(0.5f*pitch);
    sroll = sin(0.5f*roll);

    cyawcpitch = cyaw*cpitch;
    syawspitch = syaw*spitch;
    cyawspitch = cyaw*spitch;
    syawcpitch = syaw*cpitch;

    q.n = cyawcpitch * croll + syawspitch * sroll;
    q.v.x = cyawcpitch * sroll - syawspitch * croll;
    q.v.y = cyawspitch * croll + syawcpitch * sroll;
    q.v.z = syawcpitch * croll - cyawspitch * sroll;

    return q;
}

inline vector3 MakeEulerAnglesFromQ(quaternion q) {
    double r11,r21,r31,r32,r33,r12,r13;
    double q00, q11, q22, q33;
    double tmp;
    vector3 u;
    q00 = q.n * q.n;
    q11 = q.v.x * q.v.x;
    q22 = q.v.y * q.v.y;
    q33 = q.v.z * q.v.z;

    r11 = q00 + q11 - q22 - q33;
    r21 = 2 * (q.v.x*q.v.y + q.n*q.v.z);
    r31 = 2 * (q.v.x*q.v.z - q.n*q.v.y);
    r32 = 2 * (q.v.y*q.v.z + q.n*q.v.x);
    r33 = q00 - q11 - q22 + q33;

    tmp = fabs(r31);
    if(tmp > .999999) {
        r12 = 2*(q.v.x*q.v.y - q.n*q.v.z);
        r13 = 2*(q.v.x*q.v.z + q.n*q.v.y);

        u.x = RadiansToDegrees(0.0f);
        u.y = RadiansToDegrees(-(PI/2) * r31/tmp);
        u.z = RadiansToDegrees(atan2(-r12, -r31*r13));
        return u;
    }

    u.x = RadiansToDegrees(atan2(r32, r33));
    u.y = RadiansToDegrees(asin(-r31));
    u.z = RadiansToDegrees(atan2(r21, r11));
    return u;
}

inline quaternion normalize(quaternion q) {
    double m = sqrt(q.n*q.n + q.v.x*q.v.x + q.v.y*q.v.y + q.v.z*q.v.z);

    if(m <= QUATERNION_TOL) return quaternion();
    else return q/m;
}

inline matrix3x3 quaternion_to_matrix(quaternion q) {
    return matrix3x3(1-2*q.v.y*q.v.y-2*q.v.z*q.v.z, 2*q.v.x*q.v.y-2*q.n*q.v.z, 2*q.v.x*q.v.z+2*q.n*q.v.y,
                     2*q.v.x*q.v.y+2*q.n*q.v.z, 1-2*q.v.x*q.v.x-2*q.v.z*q.v.z, 2*q.v.y*q.v.z-2*q.n*q.v.x,
                     2*q.v.x*q.v.z-2*q.n*q.v.y, 2*q.v.y*q.v.z+2*q.n*q.v.x, 1-2*q.v.x*q.v.x-2*q.v.y*q.v.y);
}
#endif
